# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for importer batches
"""

from __future__ import unicode_literals, absolute_import

import logging

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler
from rattail.util import load_object


log = logging.getLogger(__name__)


class ImporterBatchHandler(BatchHandler):
    """
    Handler for importer batches.
    """
    batch_model_class = model.ImporterBatch

    def execute(self, batch, user=None, progress=None, **kwargs):
        session = orm.object_session(batch)
        metadata = sa.MetaData(schema='batch', bind=session.bind)
        row_table = sa.Table(batch.row_table, metadata, autoload=True)

        handler = load_object(batch.import_handler_spec)(config=self.config)
        handler.runas_user = user
        handler.setup()
        handler.begin_transaction()

        importer = handler.get_importer(batch.importer_key)
        assert importer
        importer.setup()

        def process(row, i):

            if row.status_code == self.enum.IMPORTER_BATCH_ROW_STATUS_CREATE:
                host_data = {}
                for col in row_table.c:
                    if col.name.startswith('key_'):
                        field = col.name[4:]
                        host_data[field] = getattr(row, col.name)
                    elif col.name.startswith('post_'):
                        field = col.name[5:]
                        host_data[field] = getattr(row, col.name)
                key = importer.get_key(host_data)
                local_object = importer.create_object(key, host_data)
                log.debug("created new %s %s: %s", importer.model_name, key, local_object)

            elif row.status_code == self.enum.IMPORTER_BATCH_ROW_STATUS_UPDATE:
                host_data = {}
                local_data = {}
                for col in row_table.c:
                    if col.name.startswith('key_'):
                        field = col.name[4:]
                        host_data[field] = getattr(row, col.name)
                        local_data[field] = getattr(row, col.name)
                    elif col.name.startswith('pre_'):
                        field = col.name[4:]
                        local_data[field] = getattr(row, col.name)
                    elif col.name.startswith('post_'):
                        field = col.name[5:]
                        host_data[field] = getattr(row, col.name)
                key = importer.get_key(host_data)
                local_object = importer.get_local_object(key)
                local_object = importer.update_object(local_object, host_data, local_data)
                log.debug("updated %s %s: %s", importer.model_name, key, local_object)

            elif row.status_code == self.enum.IMPORTER_BATCH_ROW_STATUS_DELETE:
                host_data = {}
                for col in row_table.c:
                    if col.name.startswith('key_'):
                        field = col.name[4:]
                        host_data[field] = getattr(row, col.name)
                key = importer.get_key(host_data)
                local_object = importer.get_local_object(key)
                text = six.text_type(local_object)
                importer.delete_object(local_object)
                log.debug("deleted %s %s: %s", importer.model_name, key, text)

        rows = session.query(row_table)
        self.progress_loop(process, rows, progress,
                           message="Executing import / export batch")

        importer.teardown()
        handler.teardown()
        handler.commit_transaction()
        return True



# -*- coding: utf-8; -*-
"""add email_attempt

Revision ID: ff8662bbdb53
Revises: 9f1d111fa775
Create Date: 2018-02-06 01:35:58.889529

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'ff8662bbdb53'
down_revision = u'9f1d111fa775'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # email_attempt
    op.create_table('email_attempt',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('key', sa.String(length=254), nullable=False),
                    sa.Column('sender', sa.Text(), nullable=False),
                    sa.Column('to', sa.Text(), nullable=False),
                    sa.Column('cc', sa.Text(), nullable=True),
                    sa.Column('bcc', sa.Text(), nullable=True),
                    sa.Column('subject', sa.String(length=254), nullable=False),
                    sa.Column('sent', sa.DateTime(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=False),
                    sa.Column('status_text', sa.Text(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # email_attempt
    op.drop_table('email_attempt')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Cache Helpers
"""

from __future__ import unicode_literals, absolute_import

import logging

import six
from sqlalchemy.orm import joinedload

from rattail.core import Object
from rattail.db import model


log = logging.getLogger(__name__)


class CacheKeyNotSupported(Exception):
    """
    Special error which a model cacher should raise when generating the cache
    key for a given record/object, but when it is discovered the object does
    not have sufficient data to generate the key.
    """


class ModelCacher(object):
    """
    Generic model data caching class.
    """

    def __init__(self, session, model_class, key='uuid',
                 query=None, order_by=None, query_options=None, normalizer=None,
                 expect_duplicates=False, omit_duplicates=False,
                 use_lists=False):
        self.session = session
        self.model_class = model_class
        self.key = key
        self.duplicate_keys = set()
        self.expect_duplicates = expect_duplicates
        self.omit_duplicates = omit_duplicates
        self._query = query
        self.order_by = order_by
        self.query_options = query_options
        if normalizer is None:
            self.normalize = lambda d: d
        else:
            self.normalize = normalizer
        self.use_lists = use_lists
    
    @property
    def model_name(self):
        return self.model_class.__name__

    def query(self):
        q = self._query or self.session.query(self.model_class)
        if self.order_by:
            q = q.order_by(self.order_by)
        if self.query_options:
            for option in self.query_options:
                q = q.options(option)
        return q

    def get_cache(self, progress):
        self.instances = {}
        query = self.query()
        count = query.count()
        if not count:
            return self.instances
        
        prog = None
        if progress:
            prog = progress("Caching {0} data".format(self.model_name), count)
        for i, instance in enumerate(query, 1):

            self.cache_instance(instance)

            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        if self.duplicate_keys:
            logger = log.debug if self.expect_duplicates else log.warning
            logger("found {} duplicated keys in cache".format(len(self.duplicate_keys)))
            if self.omit_duplicates:
                for key in self.duplicate_keys:
                    log.debug("removing duplicated key from cache: {}".format(repr(key)))
                    del self.instances[key]

        return self.instances

    def get_key(self, instance, normalized):
        if callable(self.key):
            return self.key(instance, normalized)
        if isinstance(self.key, six.string_types):
            return getattr(instance, self.key)
        return tuple(getattr(instance, k) for k in self.key)

    def cache_instance(self, instance):
        normalized = self.normalize(instance)
        try:
            key = self.get_key(instance, normalized)
        except CacheKeyNotSupported:
            # this means the object doesn't belong in our cache
            return
        if self.use_lists:
            self.instances.setdefault(key, []).append(normalized)
        else:
            if key not in self.instances:
                self.instances[key] = normalized
            else:
                self.duplicate_keys.add(key)
                if not self.omit_duplicates:
                    log.debug("cache already contained key, but overwriting: {}".format(repr(key)))
                    self.instances[key] = normalized


def cache_model(session, model_class, key='uuid', progress=None, **kwargs):
    """
    Convenience function for fetching a cache of data for the given model.
    """
    cacher = ModelCacher(session, model_class, key, **kwargs)
    return cacher.get_cache(progress)


class DataCacher(Object):

    def __init__(self, session=None, **kwargs):
        super(DataCacher, self).__init__(session=session, **kwargs)

    @property
    def class_(self):
        raise NotImplementedError
    
    @property
    def name(self):
        return self.class_.__name__

    def query(self):
        return self.session.query(self.class_)

    def get_cache(self, progress):
        self.instances = {}

        query = self.query()
        count = query.count()
        if not count:
            return self.instances
        
        prog = None
        if progress:
            prog = progress("Caching {0} records".format(self.name), count)

        cancel = False
        for i, instance in enumerate(query, 1):
            self.cache_instance(instance)
            if prog and not prog.update(i):
                cancel = True
                break

        if prog:
            prog.destroy()

        if cancel:
            session.close()
            return None
        return self.instances


class DepartmentCacher(DataCacher):

    class_ = model.Department

    def cache_instance(self, dept):
        self.instances[dept.number] = dept


class SubdepartmentCacher(DataCacher):

    class_ = model.Subdepartment

    def cache_instance(self, subdept):
        self.instances[subdept.number] = subdept


class CategoryCacher(DataCacher):
    class_ = model.Category

    def cache_instance(self, category):
        self.instances[category.number] = category


def cache_categories(session, progress=None):
    """
    Return a dictionary of all :class:`rattail.db.model.Category` instances,
    keyed by :attr:`number`.
    """
    cacher = CategoryCacher(session=session)
    return cacher.get_cache(progress)


class FamilyCacher(DataCacher):

    class_ = model.Family

    def cache_instance(self, family):
        self.instances[family.code] = family


class ReportCodeCacher(DataCacher):

    class_ = model.ReportCode

    def cache_instance(self, report_code):
        self.instances[report_code.code] = report_code


class BrandCacher(DataCacher):

    class_ = model.Brand

    def cache_instance(self, brand):
        self.instances[brand.name] = brand


class VendorCacher(DataCacher):

    class_ = model.Vendor

    def cache_instance(self, vend):
        self.instances[vend.id] = vend


class ProductCacher(DataCacher):

    class_ = model.Product
    with_costs = False

    def query(self):
        q = self.session.query(model.Product)
        if self.with_costs:
            q = q.options(joinedload(model.Product.costs))
            q = q.options(joinedload(model.Product.cost))
        return q

    def cache_instance(self, prod):
        self.instances[prod.upc] = prod


def get_product_cache(session, with_costs=False, progress=None):
    """
    Cache the full product set by UPC.

    Returns a dictionary of all existing products, keyed by
    :attr:`rattail.Product.upc`.
    """

    cacher = ProductCacher(session=session, with_costs=with_costs)
    return cacher.get_cache(progress)


class CustomerGroupCacher(DataCacher):

    class_ = model.CustomerGroup

    def cache_instance(self, group):
        self.instances[group.id] = group


class CustomerCacher(DataCacher):

    class_ = model.Customer

    def cache_instance(self, customer):
        self.instances[customer.id] = customer



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Data Models
"""

from __future__ import unicode_literals, absolute_import

import re

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy

from rattail.core import Object
from rattail.time import make_utc

# These are imported because most sibling modules import them *from* here (if
# that makes sense).  Probably need to think harder about the "best" place for
# these things to live.
from rattail.db.core import uuid_column, getset_factory
from rattail.db.types import GPCType


class ModelBase(Object):
    """
    Base class for all data models.
    """

    def __repr__(self):
        mapper = orm.object_mapper(self)
        keys = ['{0}={1}'.format(k.key, repr(getattr(self, k.key))) for k in mapper.primary_key]
        return "{0}({1})".format(self.__class__.__name__, ', '.join(keys))

    @classmethod
    def get_model_title(cls):
        """
        Return a "humanized" version of the model name, for display in templates etc.
        """
        if hasattr(cls, 'model_title'):
            return cls.model_title
        # convert "CamelCase" to "Camel Case"
        return re.sub(r'([a-z])([A-Z])', r'\g<1> \g<2>', cls.__name__)

    @classmethod
    def get_model_title_plural(cls):
        """
        Return a "humanized" version of the plural model name, for display in templates etc.
        """
        if hasattr(cls, 'model_title_plural'):
            return cls.model_title_plural
        return '{}s'.format(cls.get_model_title())

    @classmethod
    def make_proxy(cls, main_class, extension, name):
        setattr(main_class, name, association_proxy(extension, name,
                                                    creator=lambda value: cls(**{name: value}),
                                                    getset_factory=getset_factory))


Base = declarative_base(cls=ModelBase)


@six.python_2_unicode_compatible
class Setting(Base):
    """
    Represents a "raw" config setting stored within the database.
    """
    __tablename__ = 'setting'

    name = sa.Column(sa.String(length=255), primary_key=True)
    value = sa.Column(sa.Text())

    def __str__(self):
        return self.name or ''


class Change(Base):
    """
    Represents a changed (or deleted) record, which is pending synchronization
    to another database.
    """
    __tablename__ = 'change'

    uuid = uuid_column()

    class_name = sa.Column(sa.String(length=40), nullable=False)

    # TODO: this needs to be renamed to e.g. object_key?  now that we sometimes
    # need to (mis)use it for non-UUID keys
    instance_uuid = sa.Column(sa.String(length=255), nullable=False, doc="""
    Key for the object which was changed/etc.  In most cases this will simply
    be the object's UUID value, hence the column name.  However in some cases
    it may contain something else; this is often needed in order to properly
    sync deletions to non-rattail systems, where the UUID is meaningless.
    """)

    object_key = orm.synonym('instance_uuid', doc="""
    Synonym to the :attr:`instance_uuid` column.  New code should perhaps
    reference the attribute by this (``object_key``) name instead, as the
    ``instance_uuid`` column may be renamed at some point, to this name.
    """)

    deleted = sa.Column(sa.Boolean(), nullable=False)

    def __repr__(self):
        return "Change(class_name={}, instance_uuid={}, deleted={})".format(
            repr(self.class_name), repr(self.instance_uuid), repr(self.deleted))


class Note(Base):
    """
    Arbitrary note attached to a supported object.
    """
    __tablename__ = 'note'
    __table_args__ = (
        sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='note_fk_created_by'),
    )
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    type = sa.Column(sa.String(length=15))
    subject = sa.Column(sa.String(length=255), nullable=True)
    text = sa.Column(sa.Text(), nullable=False)
    created = sa.Column(sa.DateTime(), nullable=False, default=make_utc)

    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)
    created_by = orm.relationship('User')

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):
        return str(self.text)

    if six.PY2:
        def __unicode__(self):
            return unicode(self.text)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
File Monitor Actions
"""

from __future__ import unicode_literals, absolute_import

import os
import sys
import time
from six.moves import queue
import socket
import logging
from traceback import format_exception

from rattail.mail import send_email


log = logging.getLogger(__name__)


class Action(object):
    """
    Base class for file monitor actions.
    """

    def __init__(self, config):
        self.config = config

    def __call__(self, *args, **kwargs):
        """
        This method must be implemented in the subclass; it defines what the
        action actually *does*.  The file monitor will invoke this method for
        all new files which are discovered.
        """
        raise NotImplementedError


class StopProcessing(Exception):
    """
    Simple exception to indicate action processing should stop.  This is really
    only useful for tests.
    """


def perform_actions(profile):
    """
    Target for action threads.  Provides the main loop which checks the queue
    for new files and invokes actions for each, as they appear.
    """

    # If running on Windows, we add a step to help ensure the file is truly
    # free of competing process interests.  (In fact it would be nice to have
    # this on Linux as well, but I'm not sure how to do it there.)
    wait_for_file = lambda p: p
    if sys.platform.startswith(u'win'): # pragma: no cover
        import win32api
        from rattail.win32 import file_is_free
        def wait_for_file(path):
            while not file_is_free(path):
                win32api.Sleep(0)

    stop = False
    while not stop:

        # Suspend execution briefly, to avoid consuming so much CPU...
        time.sleep(0.01)

        try:
            path = profile.queue.get_nowait()
        except queue.Empty:
            pass
        except StopProcessing:
            stop = True
        else:
            log.debug(u"queue contained a file: {0}".format(repr(path)))

            # In some cases, processing one file may cause other related files
            # to also be processed.  When this happens, a path on the queue may
            # point to a file which no longer exists.
            if not os.path.exists(path):
                log.warning(u"file path does not exist: {0}".format(path))
                continue

            # This does nothing unless running on Windows.
            wait_for_file(path)

            for action in profile.actions:
                try:
                    invoke_action(action, path)

                except:
                    # Stop processing files altogether for this profile if it
                    # is so configured.
                    if profile.stop_on_error:
                        log.warning(u"an error was encountered, and configuration dictates that no more "
                                    u"actions will be processed for profile {0}".format(repr(profile.key)))
                        stop = True

                    # Either way no more actions should be invoked for this
                    # particular file.
                    break


def invoke_action(action, path):
    """
    Invoke a single action on a file, retrying as necessary.
    """
    attempts = 0
    errtype = None
    while True:
        attempts += 1
        log.debug(u"invoking action {0} (attempt #{1} of {2}) on file: {3}".format(
                repr(action.spec), attempts, action.retry_attempts, repr(path)))

        try:
            action.action(path, *action.args, **action.kwargs)

        except Exception as error:

            # If we've reached our final attempt, stop retrying.
            if attempts >= action.retry_attempts:
                log.debug("attempt #{} failed for action '{}' (giving up) on "
                          "file: {}".format(attempts, action.spec, path),
                          exc_info=True)
                exc_type, exc, traceback = sys.exc_info()
                send_email(action.config, 'filemon_action_error', {
                    'hostname': socket.gethostname(),
                    'path': path,
                    'action': action,
                    'attempts': attempts,
                    'error': exc,
                    'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                })
                raise

            # If this exception is not the first, and is of a different type
            # than seen previously, do *not* continue to retry.
            if errtype is not None and not isinstance(error, errtype):
                log.exception(u"new exception differs from previous one(s), giving up on "
                              u"action {0} for file: {1}".format(repr(action.spec), repr(path)))
                raise

            # Record the type of exception seen, and pause for next retry.
            log.warning(u"attempt #{0} failed for action {1} on file: {2}".format(
                    attempts, repr(action.spec), repr(path)), exc_info=True)
            errtype = type(error)
            log.debug(u"pausing for {0} seconds before making attempt #{1} of {2}".format(
                    action.retry_delay, attempts + 1, action.retry_attempts))
            if action.retry_delay:
                time.sleep(action.retry_delay)

        else:
            # No error, invocation successful.
            log.debug(u"attempt #{0} succeeded for action {1} on file: {2}".format(
                    attempts, repr(action.spec), repr(path)))
            break


def raise_exception(path, message=u"Fake error for testing"):
    """
    File monitor action which always raises an exception.

    This is meant to be a simple way to test the error handling of a file
    monitor.  For example, whether or not file processing continues for
    subsequent files after the first error is encountered.  If logging
    configuration dictates that an email should be sent, it will of course test
    that as well.
    """
    raise Exception(u'{0}: {1}'.format(message, path))


def noop(path):
    """
    File monitor action which does nothing at all.

    This exists for the sake of tests.  I doubt it's useful in any other
    context.
    """



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Utilities
"""

from __future__ import unicode_literals, absolute_import

import re
import pprint
import logging

from sqlalchemy import orm

# TODO: Deprecate/remove these imports.
from rattail.db.config import engine_from_config, get_engines, get_default_engine, configure_session


log = logging.getLogger(__name__)


class QuerySequence(object):
    """
    Simple wrapper for a SQLAlchemy (or Django, or other?) query, to make it
    sort of behave like a normal sequence, as much as needed to e.g. make an
    importer happy.
    """

    def __init__(self, query):
        self.query = query

    def __len__(self):
        try:
            return len(self.query)
        except TypeError:
            return self.query.count()

    def __iter__(self):
        return iter(self.query)


class short_session(object):
    """
    Context manager for a short-lived database session
    """

    def __init__(self, session=None, Session=None, commit=False):
        self.session = session
        self.Session = Session
        self.private_session = not bool(session)
        self.commit = commit

    def __enter__(self):
        if not self.session:
            if not self.Session:
                from rattail.db import Session
                self.Session = Session
            self.session = self.Session()
        return self.session

    def __exit__(self, exc_type, exc_value, traceback):
        if self.private_session:
            if self.commit:
                self.session.commit()
            self.session.close()
            self.session = None


def maxlen(attr):
    """
    Return the maximum length for the given attribute.
    """
    if len(attr.property.columns) == 1:
        type_ = attr.property.columns[0].type
        return getattr(type_, 'length', None)


def make_topo_sortkey(model, metadata=None):
    """
    Returns a function suitable for use as a ``key`` kwarg to a standard Python
    sorting call.  This key function will expect a single class mapper and
    return a sequence number associated with that model.  The sequence is
    determined by SQLAlchemy's topological table sorting.
    """
    if metadata is None:
        metadata = model.Base.metadata

    tables = {}
    for i, table in enumerate(metadata.sorted_tables, 1):
        tables[table.name] = i

    log.debug("topo sortkeys for '{}' will be:\n{}".format(model.__name__, pprint.pformat(
        [(i, name) for name, i in sorted(tables.items(), key=lambda t: t[1])])))

    def sortkey(name):
        mapper = orm.class_mapper(getattr(model, name))
        return tuple(tables[t.name] for t in mapper.tables)

    return sortkey


def make_full_description(brand_name, description, size):
    """
    Combine the given field values into a complete description.
    """
    fields = [
        brand_name or '',
        description or '',
        size or '']
    fields = [f.strip() for f in fields if f.strip()]
    return ' '.join(fields)


def normalize_full_name(first_name, last_name):
    """
    Normalize the given first and last name to a "full" name value.  The
    fallback return value is an empty string.
    """
    first_name = (first_name or '').strip()
    last_name = (last_name or '').strip()
    if first_name and last_name:
        return "{} {}".format(first_name, last_name)
    if first_name:
        return first_name
    if last_name:
        return last_name
    return ''


##############################
# phone number validation
##############################

class PhoneValidator(object):
    """
    Simple validator, used to ensure a phone number matches the general
    expected pattern.
    """
    # NOTE: this was stolen from FormEncode
    _phoneRE = re.compile(r'^\s*(?:1-)?(\d\d\d)[\- \.]?(\d\d\d)[\- \.]?(\d\d\d\d)(?:\s*ext\.?\s*(\d+))?\s*$', re.I)

    def __init__(self, error=False):
        self.error = error

    def validate(self, number):
        if number:
            if not self._phoneRE.search(number):
                raise ValueError("Phone number is not valid")
            return number


def validate_phone_number(number, error=False):
    """
    Validate a single phone number.
    """
    validator = PhoneValidator(error=error)
    return validator.validate(number)


def normalize_phone_number(number):
    """
    Normalize a phone number to digits only.
    """
    if number is not None:
        return re.sub(r'\D', '', number)


def format_phone_number(number):
    """
    Returns a phone number in ``(XXX) XXX-XXXX`` format if possible; otherwise
    returns the argument unaltered.
    """
    original, number = number, normalize_phone_number(number)
    if number and len(number) == 10:
        return '({}) {}-{}'.format(number[:3], number[3:6], number[6:])
    return original



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Batch Handlers
"""

from __future__ import unicode_literals, absolute_import

import os
import shutil
import warnings

from sqlalchemy import orm

from rattail.db.cache import cache_model
from rattail.time import make_utc
from rattail.util import progress_loop, load_object


class BatchHandler(object):
    """
    Base class and partial default implementation for batch handlers.  It is
    expected that all batch handlers will ultimately inherit from this base
    class, therefore it defines the implementation "interface" loosely
    speaking.  Custom batch handlers are welcome to supplement or override this
    as needed, and in fact must do so for certain aspects.

    .. attribute: populate_batches

       Simple flag to indicate whether any/all batches being handled, will
       require initial population from a relevant data source.  Note that this
       flag should be set to ``True`` if *any* batches may need population.
       Whether or not a given batch actually needs to be populated, is
       ultimately determined by the :meth:`should_populate()` method.

    .. attribute:: repopulate_when_refresh

       Flag to indicate that when a batch is refreshed, the first step of that
       should be to re-populate the batch.  The flag is ``False`` by default,
       in which case the batch is *not* repopulated, i.e. the refresh will work
       with existing batch rows.
    """
    populate_batches = False
    repopulate_when_refresh = False

    def __init__(self, config):
        self.config = config
        self.enum = config.get_enum()

    @property
    def batch_model_class(self):
        """
        Reference to the data model class of the batch type for which this
        handler is responsible, e.g. :class:`rattail.db.model.LabelBatch`.
        Each handler must define this (or inherit from one that does).
        """
        raise NotImplementedError("You must set the 'batch_model_class' attribute "
                                  "for class '{}'".format(self.__class__.__name__))

    @property
    def batch_key(self):
        """
        The "batch type key" for the handler, e.g. ``'labels'``.  This isn't
        necessarily unique among handlers, but instead refers to a unique key
        for the type of batch being handled.  The handler needn't define this,
        as it is borrowed from :attr:`batch_model_class`.
        """
        return self.batch_model_class.batch_key

    def get_model_title(self):
        return self.batch_model_class.get_model_title()

    def make_basic_batch(self, session, progress=None, **kwargs):
        """
        Make a new "basic" batch, with no customization beyond what is provided
        by ``kwargs``, which are passed directly to the batch class constructor.
        """
        kwargs.setdefault('rowcount', 0)
        kwargs.setdefault('complete', False)
        batch = self.batch_model_class(**kwargs)
        session.add(batch)
        session.flush()
        return batch

    def make_batch(self, session, progress=None, **kwargs):
        """
        Make a new batch, with initial rows if applicable.
        """
        batch = self.make_basic_batch(session, progress=progress, **kwargs)
        self.init_batch(batch, progress=progress, **kwargs)
        return batch

    def init_batch(self, batch, progress=None, **kwargs):
        """
        Initialize the batch in whatever way might make sense.  Whether this is
        required at all is up to the batch handler etc.
        """

    def add_row(self, batch, row):
        """
        (Try to) Add the given row to the given batch.  This assumes it is a
        *new* row, perhaps along with other assumptions?
        """
        session = orm.object_session(batch)
        with session.no_autoflush:
            batch.data_rows.append(row)
            self.refresh_row(row)
        if not row.removed:
            batch.rowcount += 1

    @property
    def root_datadir(self):
        """
        The absolute path of the root folder in which data for this particular
        type of batch is stored.  The structure of this path is as follows:

        .. code-block:: none

           /{root_batch_data_dir}/{batch_type_key}

        * ``{root_batch_data_dir}`` - Value of the 'batch.files' option in the
          [rattail] section of config file.
        * ``{batch_type_key}`` - Unique key for the type of batch it is.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee it.
        """
        return self.config.batch_filedir(self.batch_key)

    def datadir(self, batch):
        """
        Returns the absolute path of the folder in which the batch's source
        data file(s) resides.  Note that the batch must already have been
        persisted to the database.  The structure of the path returned is as
        follows:

        .. code-block:: none

           /{root_datadir}/{uuid[:2]}/{uuid[2:]}

        * ``{root_datadir}`` - Value returned by :meth:`root_datadir()`.
        * ``{uuid[:2]}`` - First two characters of batch UUID.
        * ``{uuid[2:]}`` - All batch UUID characters *after* the first two.

        .. note::
           While it is likely that the data folder returned by this method
           already exists, this method does not guarantee any such thing.  It
           is typically assumed that the path will have been created by a
           previous call to :meth:`make_batch()` however.
        """
        return os.path.join(self.root_datadir, batch.uuid[:2], batch.uuid[2:])

    def make_datadir(self, batch):
        """
        Returns the data folder specific to the given batch, creating if necessary.
        """
        datadir = self.datadir(batch)
        os.makedirs(datadir)
        return datadir

    # TODO: remove default attr?
    def set_input_file(self, batch, path, attr='filename'):
        """
        Assign the data file found at ``path`` to the batch.  This overwrites
        the given attribute (``attr``) of the batch and places a copy of the
        data file in the batch's data folder.
        """
        datadir = self.make_datadir(batch)
        filename = os.path.basename(path)
        shutil.copyfile(path, os.path.join(datadir, filename))
        setattr(batch, attr, filename)

    def should_populate(self, batch):
        """
        Must return a boolean indicating whether the given batch should be
        populated from an initial data source, i.e. at time of batch creation.
        Override this method if you need to inspect the batch in order to
        determine whether the populate step is needed.  Default behavior is to
        simply return the value of :attr:`populate_batches`.
        """
        return self.populate_batches

    def setup_populate(self, batch, progress=None):
        """
        Perform any setup (caching etc.) necessary for populating a batch.
        """

    def teardown_populate(self, batch, progress=None):
        """
        Perform any teardown (cleanup etc.) necessary after populating a batch.
        """

    def populate(self, batch, progress=None):
        """
        Populate the batch with initial data rows.  It is assumed that the data
        source to be used will be known by inspecting various properties of the
        batch itself.
        """
        raise NotImplementedError("Please implement `{}.populate()` method".format(batch.__class__.__name__))

    def refreshable(self, batch):
        """
        This method should return a boolean indicating whether or not the
        handler supports a "refresh" operation for the batch, given its current
        condition.  The default assumes a refresh is allowed unless the batch
        is complete or executed.

        Note that this (currently) only affects the enabled/disabled state of
        the Refresh button within the Tailbone batch view.
        """
        if batch.executed:
            return False
        if batch.complete:
            return False
        return True

    def progress_loop(self, *args, **kwargs):
        return progress_loop(*args, **kwargs)

    def setup_refresh(self, batch, progress=None):
        """
        Perform any setup (caching etc.) necessary for refreshing a batch.
        """

    def teardown_refresh(self, batch, progress=None):
        """
        Perform any teardown (cleanup etc.) necessary after refreshing a batch.
        """

    def refresh(self, batch, progress=None):
        """
        Perform a full data refresh for the batch.  What exactly this means will
        depend on the type of batch, and specific handler logic.

        Generally speaking this refresh is meant to use queries etc. to obtain
        "fresh" data for the batch (header) and all its rows.  In most cases
        certain data is expected to be "core" to the batch and/or rows, and
        such data will be left intact, with all *other* data values being
        re-calculated and/or reset etc.
        """
        session = orm.object_session(batch)
        self.setup_refresh(batch, progress=progress)
        if self.repopulate_when_refresh:
            del batch.data_rows[:]
            batch.rowcount = 0
            session.flush()
            self.populate(batch, progress=progress)
        else:
            batch.rowcount = 0

            def refresh(row, i):
                with session.no_autoflush:
                    self.refresh_row(row)
                if not row.removed:
                    batch.rowcount += 1

            self.progress_loop(refresh, batch.active_rows(), progress,
                               message="Refreshing batch data rows")
        self.refresh_batch_status(batch)
        self.teardown_refresh(batch, progress=progress)
        return True

    def refresh_row(self, row):
        """
        This method will be passed a row object which has already been properly
        added to a batch, and which has basic required fields already
        populated.  This method is then responsible for further populating all
        applicable fields for the row, based on current data within the
        relevant system(s).

        Note that in some cases this method may be called multiple times for
        the same row, e.g. once when first creating the batch and then later
        when a user explicitly refreshes the batch.  The method logic must
        account for this possibility.
        """

    def refresh_batch_status(self, batch):
        """
        Update the batch status, as needed...
        """

    def executable(self, batch):
        """
        This method should return a boolean indicating whether or not execution
        should be allowed for the batch, given its current condition.  The
        default simply returns ``True`` but you may override as needed.

        Note that this (currently) only affects the enabled/disabled state of
        the Execute button within the Tailbone batch view.
        """
        if batch is None:
            return True
        return not bool(batch.executed)

    def auto_executable(self, batch):
        """
        Must return a boolean indicating whether the given bath is eligible for
        "automatic" execution, i.e. immediately after batch is created.
        """
        return False

    def execute(self, batch, progress=None, **kwargs):
        """
        Execute the given batch, with given progress and kwargs.  That is an
        intentionally generic statement, the meaning of which must be further
        defined by the handler subclass since default is ``NotImplementedError``.
        """
        raise NotImplementedError

    def execute_many(self, batches, progress=None, **kwargs):
        """
        Execute a set of batches, with given progress and kwargs.  Default
        behavior is to simply execute each batch in succession.  Any batches
        which are already executed are skipped.
        """
        now = make_utc()
        for batch in batches:
            if not batch.executed:
                self.execute(batch, progress=progress, **kwargs)
                batch.executed = now
                batch.executed_by = kwargs['user']
        return True

    def clone(self, oldbatch, created_by):
        """
        Clone the given batch as a new batch, and return the new batch.
        """
        newbatch = self.batch_model_class()
        newbatch.created_by = created_by
        newbatch.rowcount = 0
        rowclass = newbatch.row_class
        mapper = orm.class_mapper(rowclass)

        for oldrow in oldbatch.data_rows:
            newrow = rowclass()
            for name in mapper.columns.keys():
                if name not in ('uuid', 'batch_uuid'):
                    setattr(newrow, name, getattr(oldrow, name))
            newbatch.data_rows.append(newrow)
            newbatch.rowcount += 1

        session = orm.object_session(oldbatch)
        session.add(newbatch)
        session.flush()
        return newbatch

    def cache_model(self, session, model, **kwargs):
        return cache_model(session, model, **kwargs)


def get_batch_handler(config, batch_key, default=None):
    """
    Returns a batch handler object corresponding to the given batch key.
    """
    spec = config.get('rattail.batch', '{}.handler'.format(batch_key), default=default)
    assert spec, "no handler spec could be determined for batch type: {}".format(batch_key)
    handler = load_object(spec)(config)
    return handler



# -*- coding: utf-8 -*-
"""add brand.confirmed

Revision ID: de4ebbf5bbb3
Revises: b82daacc86b7
Create Date: 2018-01-25 15:41:12.177378

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'de4ebbf5bbb3'
down_revision = u'b82daacc86b7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # brand
    op.add_column('brand', sa.Column('confirmed', sa.Boolean(), nullable=True))
    op.alter_column('brand', 'name',
               existing_type=sa.VARCHAR(length=100),
               nullable=False)
    op.create_unique_constraint(u'brand_uq_name', 'brand', ['name'])


def downgrade():

    # brand
    op.drop_constraint(u'brand_uq_name', 'brand', type_='unique')
    op.alter_column('brand', 'name',
               existing_type=sa.VARCHAR(length=100),
               nullable=True)
    op.drop_column('brand', 'confirmed')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Excel utilities
"""

from __future__ import unicode_literals, absolute_import

import datetime

import xlrd
from xlrd.xldate import xldate_as_tuple

import openpyxl
from openpyxl.styles import Font, PatternFill

from rattail.util import progress_loop


class ExcelReader(object):
    """
    Basic class for reading Excel files.
    """

    def __init__(self, path, sheet=0, sheet_name=None, header=0, datefmt='%Y-%m-%d'):
        """
        Constructor; opens an Excel file for reading.
        """
        self.book = xlrd.open_workbook(path)
        if sheet_name is not None:
            self.sheet = self.book.sheet_by_name(sheet_name)
        else:
            self.sheet = self.book.sheet_by_index(sheet)
        self.header = header
        self.fields = self.sheet.row_values(self.header)
        self.datefmt = datefmt

    def sheet_by_name(self, name):
        return self.book.sheet_by_name(name)

    def read_rows(self, progress=None):
        rows = []

        def append(row, i):
            values = self.sheet.row_values(row)
            data = dict([(self.fields[j], value)
                         for j, value in enumerate(values)])
            rows.append(data)

        progress_loop(append, range(self.header + 1, self.sheet.nrows), progress,
                      message="Reading data from Excel file")
        return rows

    def parse_date(self, value, fmt=None):
        if isinstance(value, float):
            args = xldate_as_tuple(value, self.book.datemode)
            return datetime.datetime(*args).date()
        if value:
            return datetime.datetime.strptime(value, fmt or self.datefmt).date()


class ExcelWriter(object):
    """
    Base class for Excel writers.
    """

    def __init__(self, path, fields, sheet_title=None):
        """
        Constructor; opens an Excel workbook for writing.
        """
        self.path = path
        self.fields = fields
        self.book = openpyxl.Workbook()
        self.sheet = self.book.active
        if sheet_title:
            self.sheet.title = sheet_title

    def write_header(self):
        font = Font(bold=True)
        for i, field in enumerate(self.fields, 1):
            cell = self.sheet.cell(row=1, column=i, value=field)
            cell.font = font

    def write_rows(self, rows):
        fill_even = PatternFill(patternType='solid',
                               fgColor='d9d9d9',
                               bgColor='d9d9d9')

        for i, row in enumerate(rows, 2):
            self.sheet.append(row)

            # apply row highlighting
            if i % 2 == 0:
                for j in range(1, len(self.fields) + 1):
                    cell = self.sheet.cell(row=i, column=j)
                    cell.fill = fill_even

    def auto_freeze(self, row=2, column=1):
        """
        Freeze sheet per "the usual"
        """
        self.sheet.freeze_panes = self.sheet.cell(row=row, column=column)

    def auto_filter(self):
        """
        Add auto filters for all columns.
        """
        first = self.sheet.cell(row=1, column=1)
        last = self.sheet.cell(row=self.sheet.max_row, column=self.sheet.max_column)
        cellrange = '{}:{}'.format(first.coordinate, last.coordinate)
        self.sheet.auto_filter.ref = cellrange

    def save(self):
        self.book.save(self.path)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Vendors
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.ext.associationproxy import association_proxy

from .core import Base, uuid_column, getset_factory
from .contact import PhoneNumber, EmailAddress
from .people import Person


@six.python_2_unicode_compatible
class Vendor(Base):
    """
    Represents a vendor from which products are purchased by the store.
    """
    __tablename__ = 'vendor'
    __table_args__ = (
        sa.UniqueConstraint('id', name='vendor_uq_id'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    id = sa.Column(sa.String(length=15))
    name = sa.Column(sa.String(length=50), nullable=True)
    special_discount = sa.Column(sa.Numeric(precision=5, scale=3))

    lead_time_days = sa.Column(sa.Numeric(precision=5, scale=1), nullable=True, doc="""
    The number of days expected to elapse between the order generation and receipt
    of goods.  (This description is borrowed from the SIL standard.)
    """)

    order_interval_days = sa.Column(sa.Numeric(precision=5, scale=1), nullable=True, doc="""
    The number of days expected to elapse between repeated order placement.  (This
    description is borrowed from the SIL standard.)
    """)

    def __str__(self):
        return self.name or ''

    @property
    def fax_number(self):
        for phone in self.phones:
            if 'fax' in (phone.type or '').lower():
                return phone.number

    def add_email_address(self, address, type='Info'):
        email = VendorEmailAddress(address=address, type=type)
        self.emails.append(email)

    def add_phone_number(self, number, type='Voice'):
        phone = VendorPhoneNumber(number=number, type=type)
        self.phones.append(phone)

    def get_email(self, type_='Info'):
        for email in self.emails:
            if email.type == type_:
                return email

    def get_email_address(self, *args, **kwargs):
        email = self.get_email(*args, **kwargs)
        if email:
            return email.address

    
class VendorPhoneNumber(PhoneNumber):
    """
    Represents a phone (or fax) number associated with a vendor.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


Vendor.phones = relationship(
    VendorPhoneNumber,
    backref='vendor',
    primaryjoin=VendorPhoneNumber.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorPhoneNumber.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.phone = relationship(
    VendorPhoneNumber,
    primaryjoin=sa.and_(
        VendorPhoneNumber.parent_uuid == Vendor.uuid,
        VendorPhoneNumber.preference == 1),
    foreign_keys=[VendorPhoneNumber.parent_uuid],
    uselist=False,
    viewonly=True)


class VendorEmailAddress(EmailAddress):
    """
    Represents an email address associated with a vendor.
    """

    __mapper_args__ = {'polymorphic_identity': 'Vendor'}


Vendor.emails = relationship(
    VendorEmailAddress,
    backref='vendor',
    primaryjoin=VendorEmailAddress.parent_uuid == Vendor.uuid,
    foreign_keys=[VendorEmailAddress.parent_uuid],
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorEmailAddress.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.email = relationship(
    VendorEmailAddress,
    primaryjoin=sa.and_(
        VendorEmailAddress.parent_uuid == Vendor.uuid,
        VendorEmailAddress.preference == 1),
    foreign_keys=[VendorEmailAddress.parent_uuid],
    uselist=False,
    viewonly=True)


@six.python_2_unicode_compatible
class VendorContact(Base):
    """
    Represents a point of contact (e.g. salesperson) for a vendor.
    """
    __tablename__ = 'vendor_contact'
    __table_args__ = (
        sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='vendor_contact_fk_vendor'),
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='vendor_contact_fk_person'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
    person_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)

    person = orm.relationship(
        Person,
        backref=orm.backref(
            '_vendor_contacts',
            cascade='all, delete-orphan'))

    def __str__(self):
        return str(self.person)


Vendor._contacts = relationship(
    VendorContact, backref='vendor',
    collection_class=ordering_list('preference', count_from=1),
    order_by=VendorContact.preference,
    cascade='save-update, merge, delete, delete-orphan')

Vendor.contacts = association_proxy(
    '_contacts', 'person',
    getset_factory=getset_factory,
    creator=lambda p: VendorContact(person=p))

Vendor._contact = relationship(
    VendorContact,
    primaryjoin=sa.and_(
        VendorContact.vendor_uuid == Vendor.uuid,
        VendorContact.preference == 1),
    uselist=False,
    viewonly=True)

Vendor.contact = association_proxy(
    '_contact', 'person',
    getset_factory=getset_factory)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SIL Writer
"""

from __future__ import unicode_literals, absolute_import

import datetime
from decimal import Decimal

import six

from rattail import __version__
from rattail.core import Object
from rattail.gpc import GPC
from rattail.files import temp_path
from rattail.sil import batches


__all__ = ['Writer']


class Writer(Object):

    def __init__(self, path=None, **kwargs):
        Object.__init__(self, **kwargs)
        if path is None:
            path = self.temp_path(suffix='.sil')
        self.sil_path = path
        self.fileobj = self.get_fileobj()

    def get_fileobj(self):
        return open(self.sil_path, 'wb')

    def close(self):
        self.fileobj.close()

    @classmethod
    def temp_path(cls, **kwargs):
        """
        Generate a temporary file path, using
        :func:`rattail.files.temp_path()`.
        """
        kwargs.setdefault('prefix', 'rattail.')
        kwargs.setdefault('suffix', '.sil')
        return temp_path(**kwargs)

    def consume_batch_id(self, source='RATAIL'):
        """
        Consume and return a SIL-compatible batch ID.
        """
        return batches.consume_batch_id(source)

    def val(self, value):
        """
        Returns a string version of ``value``, suitable for inclusion within a
        data row of a SIL batch.  The conversion is done as follows:

        If ``value`` is ``None``, an empty string is returned.

        If it is an ``int`` or ``decimal.Decimal`` instance, it is converted
        directly to a string (i.e. not quoted).

        If it is a ``datetime.date`` instance, it will be formatted as
        ``'%Y%j'``.

        If it is a ``datetime.time`` instance, it will be formatted as
        ``'%H%M'``.

        Otherwise, it is converted to a string if necessary, and quoted with
        apostrophes escaped.
        """
        if value is None:
            return ''
        if isinstance(value, GPC):
            return six.text_type(value)
        if isinstance(value, int):
            return six.text_type(value)
        if isinstance(value, float):
            return six.text_type(value)
        if isinstance(value, Decimal):
            return six.text_type(value)
        if isinstance(value, datetime.date):
            return value.strftime('%Y%j')
        if isinstance(value, datetime.time):
            return value.strftime('%H%M')
        if not isinstance(value, six.string_types):
            value = six.text_type(value)
        return "'{}'".format(value.replace("'", "''"))

    def write(self, string):
        self.fileobj.write(string)

    def write_batch_header_raw(self, **kwargs):
        """
        Writes a SIL batch header string.  All keyword arguments correspond to
        the SIL specification for the Batch Header Dictionary.

        **Batch Header Dictionary:**

        ====  ====    ====  ===========
        Name  Type    Size  Description
        ====  ====    ====  ===========
        H01   CHAR       2  Batch Type
        H02   CHAR       8  Batch Identifier
        H03   CHAR       6  Source Identifier
        H04   CHAR       6  Destination Identifier
        H05   CHAR      12  Audit File Name
        H06   CHAR      12  Response File Name
        H07   DATE       7  Origin Date
        H08   TIME       4  Origin Time
        H09   DATE       7  Execution (Apply) Date
        H10   DATE       4  Execution (Apply) Time
        H11   DATE       7  Purge Date
        H12   CHAR       6  Action Type
        H13   CHAR      50  Batch Description
        H14   CHAR      30  User Defined
        H15   CHAR      30  User Defined
        H16   CHAR      30  User Defined
        H17   NUMBER     1  Warning Level
        H18   NUMBER     5  Maximum Error Count
        H19   CHAR       7  SIL Level/Revision
        H20   CHAR       4  Software Revision
        H21   CHAR      50  Primary Key
        H22   CHAR     512  System Specific Command
        H23   CHAR       8  Dictionary Revision

        Consult the SIL Specification for more information.
        """

        kw = kwargs

        # Don't quote H09 if special "immediate" value.
        H09 = kw.get('H09')
        if H09 != '0000000':
            H09 = self.val(H09)

        # Don't quote H10 if special "immediate" value.
        H10 = kw.get('H10')
        if H10 != '0000':
            H10 = self.val(H10)

        row = [
            self.val(kw.get('H01')),
            self.val(kw.get('H02')),
            self.val(kw.get('H03')),
            self.val(kw.get('H04')),
            self.val(kw.get('H05')),
            self.val(kw.get('H06')),
            self.val(kw.get('H07')),
            self.val(kw.get('H08')),
            H09,
            H10,
            self.val(kw.get('H11')),
            self.val(kw.get('H12')),
            self.val(kw.get('H13')),
            self.val(kw.get('H14')),
            self.val(kw.get('H15')),
            self.val(kw.get('H16')),
            self.val(kw.get('H17')),
            self.val(kw.get('H18')),
            self.val(kw.get('H19')),
            self.val(kw.get('H20')),
            self.val(kw.get('H21')),
            self.val(kw.get('H22')),
            self.val(kw.get('H23')),
            ]

        self.write('INSERT INTO HEADER_DCT VALUES\n')
        self.write_row(row, quote=False, last=True)
        self.write('\n')

    def write_batch_header(self, **kwargs):
        """
        Convenience method to take some of the gruntwork out of writing batch
        headers.

        If you do not override ``H03`` (Source Identifier), then Rattail will
        provide a default value for it, as well as ``H20`` (Software Revision)
        - that is, unless you've supplied it yourself.

        If you do not provide values for ``H07`` or ``H08``, the current date
        and time will be assumed.

        If you do not provide values for ``H09`` or ``H10``, it is assumed that
        you wish the batch to be immediately executable.  Default values will
        be provided accordingly.

        If you do not provide a value for ``H11`` (Purge Date), a default of 90
        days from the current date will be assumed.
        """

        kw = kwargs

        # Provide default for H03 (Source Identifier) if none specified.
        if 'H03' not in kw:
            kw['H03'] = 'RATAIL'

            # Provide default for H20 (Software Revision) if none specified.
            if 'H20' not in kw:
                kw['H20'] = __version__[:4]

        # Provide default (current local time) values H07 and H08 (Origin Date /
        # Time) if none was specified.
        now = datetime.datetime.now()
        if 'H07' not in kw:
            kw['H07'] = now.date()
        if 'H08' not in kw:
            kw['H08'] = now.time()

        # Use special "immediate" values for H09 and H10 (Execution (Apply) Date /
        # Time) if none was specified.
        if 'H09' not in kw:
            kw['H09'] = '0000000'
        if 'H10' not in kw:
            kw['H10'] = '0000'

        # Provide default value for H11 (Purge Date) if none was specified.
        if 'H11' not in kw:
            kw['H11'] = (now + datetime.timedelta(days=90)).date()

        self.write_batch_header_raw(**kw)

    def write_create_header(self, **kwargs):
        """
        Convenience method to take some of the gruntwork out of writing batch
        headers.

        The following default values are provided by this method:

        * ``H01`` = ``'HC'``
        * ``H12`` = ``'LOAD'``

        This method also calls :meth:`write_batch_header()`; see its
        documentation for the other default values provided.
        """

        kw = kwargs
        kw.setdefault('H01', 'HC')
        kw.setdefault('H12', 'LOAD')
        self.write_batch_header(**kw)

    def write_maintenance_header(self, **kwargs):
        """
        Convenience method to take some of the gruntwork out of writing batch
        headers.

        The following default values are provided by this method:

        * ``H01`` = ``'HM'``

        This method also calls :meth:`write_batch_header()`; see its
        documentation for the other default values provided.
        """

        kw = kwargs
        kw.setdefault('H01', 'HM')
        self.write_batch_header(**kw)

    def write_row(self, row, quote=True, last=False):
        """
        Writes a SIL row string.

        ``row`` should be a sequence of values.

        If ``quote`` is ``True``, each value in ``row`` will be ran through the
        :func:`val()` function before being written.  If it is ``False``, the
        values are written as-is.

        If ``last`` is ``True``, then ``';'`` will be used as the statement
        terminator; otherwise ``','`` is used.
        """

        terminator = ';' if last else ','
        if quote:
            row = [self.val(x) for x in row]
        self.write('(' + ','.join(row) + ')' + terminator + '\n')

    def write_rows(self, rows, count=None, progress=None,
                  progress_text="Writing data to SIL file"):
        """
        Write some SIL data row strings to the file object.

        This is a wrapper around :meth:`write_row()` which adds some
        convenience, and hopefully improves efficiency.  It should make calling
        code a little prettier if nothing else.

        :param rows: Provider of data rows, each of which should be a sequence
          of data suitable for use with :meth:`write_row()`.  ``rows`` itself
          may be a sequence or a generator, or a callable which returns either
          of these.

        :param count: Optional count of ``rows``, in case ``len(rows)`` doesn't
          give the right answer or is expensive to calculate.

        :param progress: Optional progress indicator factory.

        :returns: ``True``, unless the operation is canceled by the user.
        """
        if callable(rows):
            rows = rows()
        if count is None:
            count = len(rows)

        prog = None
        if progress:
            prog = progress(progress_text, count)

        cancel = False
        for i, row in enumerate(rows, 1):
            self.write_row(row, last=i == count)
            if prog and not prog.update(i):
                cancel = True
                break

        if prog:
            prog.destroy()
        return not cancel



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for pricing batches
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler


class PricingBatchHandler(BatchHandler):
    """
    Handler for pricing batches.
    """
    batch_model_class = model.PricingBatch

    def populate(self, batch, progress=None):
        """
        Batch row data comes from product query.
        """
        assert batch.products
        session = orm.object_session(batch)

        def append(item, i):
            row = model.PricingBatchRow()
            row.product = item
            row.upc = row.product.upc
            self.add_row(batch, row)

        assert self.progress_loop(append, batch.products, progress,
                                  message="Adding initial rows to batch")

    def refresh_row(self, row):
        """
        Inspect a row from the source data and populate additional attributes
        for it, according to what we find in the database.
        """
        product = row.product
        assert product

        row.brand_name = six.text_type(product.brand or '')
        row.description = product.description
        row.size = product.size
        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None

        cost = product.cost
        row.vendor = cost.vendor if cost else None
        row.regular_unit_cost = cost.unit_cost if cost else None

        price = product.regular_price
        row.old_price = price.price if price else None

    def set_status_per_diff(self, row):
        """
        Set the row's status code according to its price diff
        """
        threshold = row.batch.min_diff_threshold
        minor = bool(threshold) and abs(row.price_diff) < threshold
        if row.price_diff > 0:
            if minor:
                row.status_code = row.STATUS_PRICE_INCREASE_MINOR
            else:
                row.status_code = row.STATUS_PRICE_INCREASE
        elif row.price_diff < 0:
            if minor:
                row.status_code = row.STATUS_PRICE_DECREASE_MINOR
            else:
                row.status_code = row.STATUS_PRICE_DECREASE
        else:
            row.status_code = row.STATUS_PRICE_UNCHANGED



# -*- coding: utf-8 -*-
"""add product.last_sold

Revision ID: 3623367c499f
Revises: 6cadbeb3f82
Create Date: 2015-03-10 16:02:27.105206

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3623367c499f'
down_revision = u'6cadbeb3f82'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product.last_sold
    op.add_column('product', sa.Column('last_sold', sa.DateTime(), nullable=True))
    op.add_column('product_version', sa.Column('last_sold', sa.DateTime(), nullable=True))


def downgrade():

    # product.last_sold
    op.drop_column('product_version', 'last_sold')
    op.drop_column('product', 'last_sold')



# -*- coding: utf-8 -*-
"""add datasync change

Revision ID: 188810a5f9db
Revises: 3a7029aed67
Create Date: 2015-10-17 01:07:39.015318

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '188810a5f9db'
down_revision = u'3a7029aed67'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.create_table('datasync_change',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('source', sa.String(length=20), nullable=False),
                    sa.Column('payload_type', sa.String(length=20), nullable=False),
                    sa.Column('payload_key', sa.String(length=50), nullable=False),
                    sa.Column('deletion', sa.Boolean(), nullable=False),
                    sa.Column('obtained', sa.DateTime(), nullable=False),
                    sa.Column('consumer', sa.String(length=20), nullable=True),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():
    op.drop_table('datasync_change')



# -*- coding: utf-8 -*-
"""use decimal qty for handheld/inventory rows

Revision ID: a47a4d73b662
Revises: 0f4cb39da3cc
Create Date: 2016-12-16 16:12:48.505269

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'a47a4d73b662'
down_revision = u'0f4cb39da3cc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # batch_handheld_row
    op.alter_column('batch_handheld_row', 'cases', type_=sa.Numeric(precision=10, scale=4))
    op.alter_column('batch_handheld_row', 'units', type_=sa.Numeric(precision=10, scale=4))

    # batch_inventory_row
    op.alter_column('batch_inventory_row', 'cases', type_=sa.Numeric(precision=10, scale=4))
    op.alter_column('batch_inventory_row', 'units', type_=sa.Numeric(precision=10, scale=4))


def downgrade():

    # batch_inventory_row
    op.alter_column('batch_inventory_row', 'cases', type_=sa.Integer())
    op.alter_column('batch_inventory_row', 'units', type_=sa.Integer())

    # batch_handheld_row
    op.alter_column('batch_handheld_row', 'cases', type_=sa.Integer())
    op.alter_column('batch_handheld_row', 'units', type_=sa.Integer())



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for label batches
"""

from __future__ import unicode_literals, absolute_import

import csv
import decimal
import logging

import six
import json
import sqlalchemy as sa
from sqlalchemy import orm

from rattail import enum
from rattail.db import model, api
from rattail.gpc import GPC
from rattail.batch import BatchHandler
from rattail.csvutil import UnicodeDictReader
from rattail.util import progress_loop
from rattail.time import make_utc
from rattail.config import parse_bool


log = logging.getLogger(__name__)


class LabelBatchHandler(BatchHandler):
    """
    Handler for Print Labels batches.
    """
    batch_model_class = model.LabelBatch

    def setup(self, batch, progress=None):
        self.now = make_utc()

    setup_populate = setup
    setup_refresh = setup

    def make_batch(self, session, progress=None, **kwargs):
        """
        Make a new batch, with initial rows if applicable.
        """
        self.skip_first_line = parse_bool(kwargs.pop('skip_first_line', False))
        self.calc_check_digit = kwargs.pop('calc_check_digit', False)
        if self.calc_check_digit != 'upc':
            self.calc_check_digit = parse_bool(self.calc_check_digit)
        file_has_options = parse_bool(kwargs.pop('file_has_options', False))
        batch = super(LabelBatchHandler, self).make_batch(session, progress, **kwargs)
        batch.file_has_options = file_has_options
        return batch

    def auto_executable(self, batch):
        """
        Must return a boolean indicating whether the given bath is eligible for
        "automatic" execution, i.e. immediately after batch is created.
        """
        if batch.filename and '.autoexecute.' in batch.filename:
            return True
        return False

    def populate(self, batch, progress=None):
        """
        Pre-fill batch with row data from handheld batch, etc.
        """
        assert batch.handheld_batch or batch.filename or batch.products
        session = orm.object_session(batch)
        self.label_profile = self.get_label_profile(session)
        self.setup_populate(batch, progress=progress)

        def append(item, i):
            row = model.LabelBatchRow()
            row.label_code = self.label_profile.code if self.label_profile else None
            row.label_profile = self.label_profile
            with session.no_autoflush:
                if isinstance(item, model.Product):
                    row.product = item
                    row.upc = row.product.upc
                    row.label_quantity = 1
                    if batch.static_prices and hasattr(item, '_batch_price'):
                        row.regular_price = item._batch_price
                else: # item is handheld batch row
                    row.upc = item.upc
                    row.product = item.product
                    row.label_quantity = item.units or 1
                    # copy these in case product is null
                    row.brand_name = item.brand_name
                    row.description = item.description
                    row.size = item.size
                batch.add_row(row)
            self.refresh_row(row)

        if batch.handheld_batch:
            data = batch.handheld_batch.active_rows()
        elif batch.filename:
            if batch.file_has_options:
                self.set_options_from_file(batch)
            data = self.read_products_from_file(batch, progress=progress)
        elif batch.products:
            data = batch.products

        self.progress_loop(append, data, progress,
                           message="Adding initial rows to batch")

    def set_options_from_file(self, batch):
        """
        Set various batch options, if any are present within the data file.
        """
        path = batch.filepath(self.config)
        with open(path, 'rb') as f:
            options = json.loads(f.readline())
        if 'description' in options and options['description']:
            batch.description = options['description']
        if 'notes' in options and options['notes']:
            batch.notes = options['notes']
        if 'static_prices' in options:
            batch.static_prices = options['static_prices']

    def read_products_from_file(self, batch, progress=None):
        """
        Returns list of Product objects based on lookup from CSV file data.

        # TODO: should this actually happen here? vs refresh and just mark product not found?
        """
        path = batch.filepath(self.config)
        with open(path, 'rb') as f:
            if self.skip_first_line:
                f.readline()
                reader = csv.reader(f)
                data = [{'upc': row[0]} for row in reader]
            else:
                fields = None
                if batch.file_has_options:
                    f.readline()
                    reader = csv.reader(f)
                    fields = reader.next()
                    f.seek(0)
                    f.readline()
                    f.readline()
                reader = UnicodeDictReader(f, fieldnames=fields)
                data = list(reader)

        products = []
        session = orm.object_session(batch)

        def append(entry, i):
            upc = entry['upc'].strip()
            if upc:
                upc = GPC(upc, calc_check_digit=self.calc_check_digit)
                product = api.get_product_by_upc(session, upc)
                if product:
                    if batch.static_prices and entry['regular_price']:
                        product._batch_price = decimal.Decimal(entry['regular_price'])
                    products.append(product)
                else:
                    log.warning("product not found: {}".format(upc))

        self.progress_loop(append, data, progress,
                           message="Reading data from CSV file")
        return products

    def get_label_profile(self, session):
        code = self.config.get('rattail.batch', 'labels.default_code')
        if code:
            return session.query(model.LabelProfile)\
                          .filter(model.LabelProfile.code == code)\
                          .one()
        else:
            return session.query(model.LabelProfile)\
                          .order_by(model.LabelProfile.ordinal)\
                          .first()

    def refresh_row(self, row):
        """
        Inspect a row from the source data and populate additional attributes
        for it, according to what we find in the database.
        """
        if not row.product and row.upc:
            session = orm.object_session(row)
            row.product = api.get_product_by_upc(session, row.upc)
        if not row.product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        product = row.product
        row.brand_name = six.text_type(product.brand or '')
        row.description = product.description
        row.size = product.size
        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None
        category = product.category
        row.category_code = category.code if category else None
        row.category_name = category.name if category else None
        if not row.batch.static_prices:
            regular_price = product.regular_price
            row.regular_price = regular_price.price if regular_price else None
            row.pack_quantity = regular_price.pack_multiple if regular_price else None
            row.pack_price = regular_price.pack_price if regular_price else None
            sale_price = product.current_price
            if sale_price:
                if (sale_price.type == enum.PRICE_TYPE_SALE and
                    sale_price.starts and sale_price.starts <= self.now and
                    sale_price.ends and sale_price.ends >= self.now):
                    pass            # this is what we want
                else:
                    sale_price = None
            row.sale_price = sale_price.price if sale_price else None
            row.sale_start = sale_price.starts if sale_price else None
            row.sale_stop = sale_price.ends if sale_price else None
        cost = product.cost
        vendor = cost.vendor if cost else None
        row.vendor_id = vendor.id if vendor else None
        row.vendor_name = vendor.name if vendor else None
        row.vendor_item_code = cost.code if cost else None
        row.case_quantity = cost.case_size if cost else None
        if row.regular_price:
            row.status_code = row.STATUS_OK
        else:
            row.status_code = row.STATUS_REGULAR_PRICE_UNKNOWN

    def execute(self, batch, progress=None, **kwargs):
        """
        Print some labels!
        """
        return self.print_labels(batch, progress)

    def print_labels(self, batch, progress=None):
        """
        Print all labels for the given batch.
        """
        profiles = {}

        def organize(row, i):
            profile = row.label_profile
            if profile.uuid not in profiles:
                profiles[profile.uuid] = profile
                profile.labels = []

            # TODO: should include more things here?
            data = {}
            if batch.static_prices:
                data['price'] = row.regular_price

            profile.labels.append((row.product, row.label_quantity, data))

        # filter out removed rows, and maybe inactive product rows
        rows = batch.active_rows()
        if self.config.getbool('rattail.batch', 'labels.exclude_inactive_products', default=False):
            rows = [row for row in rows
                    if row.status_code not in (row.STATUS_PRODUCT_APPEARS_INACTIVE,
                                               row.STATUS_PRODUCT_NOT_FOUND)]

        progress_loop(organize, rows, progress,
                      message="Organizing labels by type")

        # okay now print for real
        # TODO: this is compatible with legacy code but will re-read all needed
        # product data from master table instead of levaraging batch rows
        for profile in profiles.values():
            printer = profile.get_printer(self.config)
            printer.print_labels(profile.labels, progress=progress)

        return True



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Utilities
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm


def make_dependency_sorter(model):
    """
    Return a function suitable for use when sorting data model class names,
    according to the underlying foreign key dependencies between the tables.
    """

    def depends_on(mapper_x, mapper_y, visited=set()):
        """
        Returns True if model 'x' has a (in)direct FK dependency on 'y', else False.
        """
        # First check for direct FK dependency.
        for column in mapper_x.columns:
            for fkey in column.foreign_keys:
                if fkey.column.table in mapper_y.tables:
                    return True

        # Must keep track of where we're been, to avoid infinite recursion in
        # the case of mutually-dependent tables.
        visited.add(mapper_x)

        # Next check recursively, for indirect FK dependency.
        for prop in mapper_x.iterate_properties:
            if isinstance(prop, orm.RelationshipProperty):
                if prop.direction.name == 'MANYTOONE':
                    if prop.mapper not in visited:
                        if depends_on(prop.mapper, mapper_y, visited=visited):
                            return True

        return False

    def dependency_sorter(x, y):
        """
        This function is meant to be used as the ``cmp`` kwarg to a standard
        Python sorting function.
        """
        map_x = orm.class_mapper(getattr(model, x))
        map_y = orm.class_mapper(getattr(model, y))

        # Check to see if one model depends (in)directly on the other.  If "x
        # depends on y" then we consider "x > y" and hence the final sort
        # result would be (y, x).  Note however that in the case of mutually
        # dependent tables, we must remain neutral and allow the tie.
        dep_x = depends_on(map_x, map_y)
        dep_y = depends_on(map_y, map_x)
        if dep_x and not dep_y:
            return 1
        if dep_y and not dep_x:
            return -1
        return 0

    return dependency_sorter



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
File monitor actions for creating batches
"""

from __future__ import unicode_literals, absolute_import

import logging

from rattail.db import Session, model
from rattail.config import parse_bool
from rattail.filemon import Action
from rattail.time import make_utc
from rattail.util import load_object


log = logging.getLogger(__name__)


class MakeBatch(Action):
    """
    Filemon action for making new file-based batches.  Note that this only
    handles the simple case of creating a batch with a single input data file.
    """

    def __call__(self, path, batch_type='', handler='', user='',
                 delete_if_empty='false', auto_execute_allowed='true',
                 **kwargs):
        """
        Make a batch from the given file path.

        :param path: Path to the source data file for the batch.

        :param type: Type of batch, specified by the batch "key" - e.g. 'labels'.

        :param handler: Spec string for the batch handler class, e.g.
           ``'rattail.batch.handheld.handler:HandheldBatchHandler'``.  Note
           that this argument is *required*.

        :param user: Username of the user which is to be credited with creating
           the batch.

        :param delete_if_empty: Set this to 'true' if you wish the batch to be
           auto-deleted in the event that it has no data rows, after being
           initially populated from the given data file.

        :param auto_execute_allowed: Set this to 'false' if you wish to disable
           auto-execution of new batches.  Default behavior is for the handler
           to decide whether each batch is eligible for auto-execution.  Note
           that it is not possible to "force" auto-execution of a batch; you
           may only "allow" or "prevent" it.
        """
        if not handler:
            assert batch_type
            handler = self.config.require('rattail.batch', '{}.handler'.format(batch_type))
        handler = load_object(handler)(self.config)
        session = Session()
        user = session.query(model.User).filter_by(username=user).one()
        kwargs['created_by'] = user
        batch = handler.make_batch(session, **kwargs)
        handler.set_input_file(batch, path)
        handler.populate(batch)
        handler.refresh_batch_status(batch)

        if parse_bool(delete_if_empty):
            session.flush()
            if not batch.data_rows:
                log.debug("auto-deleting empty '{}' batch: {}".format(handler.batch_key, batch))
                batch.delete_data(self.config)
                session.delete(batch)
                batch = None

        if batch and parse_bool(auto_execute_allowed) and handler.auto_executable(batch):
            handler.execute(batch, user=user)
            batch.executed = make_utc()
            batch.executed_by = user

        session.commit()
        session.close()



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase
from six import StringIO

from mock import patch

from rattail import csvutil


class TestDictWriter(TestCase):

    def test_writeheader_26(self):
        # Simulate Python 2.6
        with patch('csv.writer'):
            with patch('rattail.csvutil.csv.DictWriter', spec=['writer']) as DictWriter:
                buf = StringIO()
                writer = csvutil.DictWriter(buf, ['field1', 'field2'])
                writer.writeheader()
                buf.close()
                writer.writer.writerow.assert_called_once_with(['field1', 'field2'])

    def test_writeheader_27(self):
        # Simulate Python 2.7+
        with patch('csv.writer'):
            with patch('rattail.csvutil.csv.DictWriter', spec=['writer', 'writeheader']) as DictWriter:
                buf = StringIO()
                writer = csvutil.DictWriter(buf, ['field1', 'field2'])
                writer.writeheader()
                buf.close()
                self.assertFalse(writer.writer.writerow.called)
                DictWriter.writeheader.assert_called_once_with(writer)
